/*
  This is a fairly straightforward implementation of a B-Tree in memory modeled after 
  Cormen, et al., Introduction to Algorithms, Chapter 18
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include "btree_mem.h"

struct btree_mem_node
   {
   int nkeys;
   void ** key;
   void ** data;
   struct btree_mem_node ** child;
   bool leaf;
   };

struct btree_mem
   {
   int mindeg;
   struct btree_mem_node * root;
   int (*cmp)(void *, void *);
   };

struct btree_mem_node * btree_mem_node_new            (struct btree_mem * bt);
void                    btree_mem_node_delete         (struct btree_mem_node * node);

struct btree_mem_node * btree_mem_node_find           (struct btree_mem * bt, struct btree_mem_node * node, void * key, int * i);

void *                  btree_mem_node_insert         (struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data);
void *                  btree_mem_node_insert_nonfull (struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data);

struct btree_mem_node * btree_mem_split_child         (struct btree_mem * bt, struct btree_mem_node * x, int i);

void                    btree_mem_node_walk_inorder   (struct btree_mem_node * node, void (*f)(int, void const * const, void *), int depth);
void                    btree_mem_node_dump           (struct btree_mem_node * node, void (*f)(int, void const * const, void *), int depth);

void *                  btree_mem_node_remove         (struct btree_mem * bt, struct btree_mem_node * node, void * key, 
						       void ** key_value, void ** data_value);

void                    btree_mem_node_add_key        (struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data);
void *                  btree_mem_node_remove_key     (struct btree_mem_node * node, int i);
void *                  btree_mem_node_first_key      (struct btree_mem_node * node);
void *                  btree_mem_node_last_key       (struct btree_mem_node * node);
void                    btree_mem_node_merge          (struct btree_mem_node * a, struct btree_mem_node * b);

#ifdef DEBUG
#define DEBUG_BLOCK(x) x

void btree_mem_nonde_print_keys(struct btree_mem_node * node);

#else
#define DEBUG_BLOCK(x)
#endif

void printit(int h, void * k, void * d)
   {
   for (int i = 0; i < h; i++)
      printf("\t");
   printf("%d %s\n", *((int*)k), (char *) d);
   }


////////////////////////////////////////////////////////////////////////////////
struct btree_mem *
btree_mem_new(int min_degree, int (*cmp)(void *, void *))
   {
   struct btree_mem * bt = malloc(sizeof(struct btree_mem));
   if (NULL == bt)
      return NULL;

   bt->mindeg = min_degree;
   bt->cmp = cmp;

   bt->root = btree_mem_node_new(bt);
   if (NULL == bt->root)
      {
      free(bt);
      return NULL;
      }

   bt->root->leaf = true;
   return bt;
   }

////////////////////////////////////////////////////////////////////////////////
// IMPORTANT NOTE: All keys and data must be removed first or else memory
// will leak.
void
btree_mem_delete(struct btree_mem * bt)
   {
   // TODO: walk down the tree and delete all nodes.
   free(bt);
   }

////////////////////////////////////////////////////////////////////////////////
struct btree_mem_node *
btree_mem_node_new(struct btree_mem * bt)
   {
   struct btree_mem_node * node = malloc(sizeof(struct btree_mem_node));
   if (NULL == node)
      return NULL;

   int maxchild = 2*bt->mindeg;

   node->key  = malloc( (maxchild - 1) * sizeof(void *));
   if (NULL == node->key)
      {
      free(node);
      return NULL;
      }

   node->data = malloc( (maxchild - 1) * sizeof(void *));
   if (NULL == node->data)
      {
      free(node->key);
      free(node);
      return NULL;
      }

   node->child = malloc( maxchild * sizeof(struct btree_mem_node *));
   if (NULL == node->data)
      {
      free(node->data);
      free(node->key);
      free(node);
      return NULL;
      }

   for (int i = 0; i < maxchild - 1; i++)
      {
      node->key[i] = NULL;
      node->data[i] = NULL;
      node->child[i] = NULL;
      }
   node->child[maxchild-1] = NULL;

   node->nkeys = 0;
   node->leaf = false;

   return node;
   }

////////////////////////////////////////////////////////////////////////////////
// IMPORTANT NOTE: All keys and data must be removed first or else memory
// will leak.
void
btree_mem_node_delete(struct btree_mem_node * node)
   {
   free(node->child);
   free(node->data);
   free(node->key);
   free(node);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_find(struct btree_mem * bt, void * key)
   {
   int i;
   struct btree_mem_node * node = btree_mem_node_find(bt, bt->root, key, &i);
   if (NULL == node)
      return NULL;
   return node->data[i];
   }

////////////////////////////////////////////////////////////////////////////////
struct btree_mem_node *
btree_mem_node_find(struct btree_mem * bt, struct btree_mem_node * node, void * key, int * i)
   {
   *i = 0;
   while (( *i < node->nkeys) && (bt->cmp(key, node->key[*i]) > 0) )
      (*i)++;
   if ( (*i < node->nkeys) && (bt->cmp(key, node->key[*i]) == 0) )
      return node;
   if (node->leaf)
      return NULL;
   else
      return btree_mem_node_find(bt, node->child[*i], key, i);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_insert(struct btree_mem * bt, void * key, void * data)
   {
   DEBUG_BLOCK(  printf("Inserting key %d data >%s<\n", *((int *) key), (char *) data);  )
   struct btree_mem_node * s = bt->root;

   // If the root node is full, split it here.
   if ((2 * bt->mindeg - 1) == s->nkeys)
      {
      s = btree_mem_node_new(bt);
      if (NULL == s)
	 return NULL;
      s->child[0] = bt->root;
      bt->root = s;
      btree_mem_split_child(bt, s, 0);
      }
   return btree_mem_node_insert_nonfull(bt, s, key, data);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_node_insert_nonfull(struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data)
   {
   int i = node->nkeys - 1;
   if (node->leaf)
      {
      //DEBUG_BLOCK( printf("node->nkeys: %d\n", node->nkeys); )
      while ( (i >= 0) && (bt->cmp(key, node->key[i]) < 0) )
	 {
	 node->key[i+1] = node->key[i];
	 node->data[i+1] = node->data[i];
	 i--;
	 }
      node->key[i+1] = key;
      node->data[i+1] = data;
      node->nkeys++;
      return key;
      }
   else
      {
      while ( (i >= 0) && (bt->cmp(key, node->key[i]) < 0) )
	 i--;
      i++;

      if (node->child[i]->nkeys == (2 * bt->mindeg - 1))
	 {
	 if (NULL == btree_mem_split_child(bt, node, i))
	    return NULL;
	 if (bt->cmp(key, node->key[i]) > 0)
	    i++;
	 }
      return btree_mem_node_insert_nonfull(bt, node->child[i], key, data);
      }
   }

////////////////////////////////////////////////////////////////////////////////
// x - a non-full node
// i - The index of a full node y in x's child list.
//
// Split y in half, adding the newly created node z as a new child of x.
// z will be placed after y in the child list.

struct btree_mem_node *
btree_mem_split_child(struct btree_mem * bt, struct btree_mem_node * x, int i)
   {

   struct btree_mem_node * y = x->child[i];

   // Allocate new node.
   struct btree_mem_node * z = btree_mem_node_new(bt);
   if (NULL == z)
      return NULL;

   // Load z with the second half of y's data;
   z->leaf = y->leaf;
   z->nkeys = bt->mindeg - 1;
   for (int j = 0; j < bt->mindeg - 1; j++)
      {
      z->key[j] = y->key[j + bt->mindeg];
      z->data[j] = y->data[j + bt->mindeg];
      }
   if (! y->leaf)
      {
      for (int j = 0; j < bt->mindeg; j++)
	 z->child[j] = y->child[j + bt->mindeg];
      }

   // Update y's nkeys.
   y->nkeys = bt->mindeg - 1;

   // Make a hole in x's child list for the new node z.
   for (int j = x->nkeys; j >= i + 1; j--)
      x->child[j+1] = x->child[j];
   x->child[i+1] = z;

   // Make a hole in x's key list for the new node z.
   for (int j = x->nkeys - 1; j >= i; j--)
      {
      x->key[j+1] = x->key[j];
      x->data[j+1] = x->data[j];
      }
   x->key[i] = y->key[bt->mindeg-1];
   x->data[i] = y->data[bt->mindeg-1];

   x->nkeys += 1;

   return z;
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_walk_inorder(struct btree_mem * bt, void (*f)(int, void const * const, void *))
   {
   btree_mem_node_walk_inorder(bt->root, f, 0);
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_node_walk_inorder(struct btree_mem_node * node, void (*f)(int, void const * const, void *), int depth)
   {

   if (0 == node->nkeys)
      {
      if (NULL != node->child[0])
	 btree_mem_node_walk_inorder(node->child[0], f, depth+1);
      return;
      }

   for (int i = 0; i < node->nkeys; i++)
      {
      if (! node->leaf) 
	 btree_mem_node_walk_inorder(node->child[i], f, depth+1);
      f(depth, node->key[i], node->data[i]);
      }
   if (! node->leaf) 
      btree_mem_node_walk_inorder(node->child[node->nkeys], f, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_dump(struct btree_mem * bt, void (*f)(int, void const * const, void *))
   {
   btree_mem_node_dump(bt->root, f, 0);
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_node_dump(struct btree_mem_node * node, void (*f)(int, void const * const, void *), int depth)
   {

   if (0 == node->nkeys)
      {
      if (NULL != node->child[0])
	 btree_mem_node_dump(node->child[0], f, depth+1);
      return;
      }

   for (int i = 0; i < node->nkeys; i++)
      {
      if (! node->leaf) 
	 btree_mem_node_dump(node->child[i], f, depth+1);
      printf("node: %p nkeys %10d \t", node, node->nkeys);
      f(depth, node->key[i], node->data[i]);
      }
   if (! node->leaf) 
      btree_mem_node_dump(node->child[node->nkeys], f, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_remove(struct btree_mem * bt, void * key, void ** key_value, void ** data_value)
   {
   if ( (NULL == bt) || (NULL == key) || (NULL == key_value) || (NULL == data_value) )
      return NULL;

   DEBUG_BLOCK( printf("Removing %d\n", *((int *) key)); )
   void * data = btree_mem_node_remove(bt, bt->root, key, key_value, data_value);
   if (NULL == data)
      return NULL;

   if (0 == bt->root->nkeys)
      {
      DEBUG_BLOCK( printf("Freeing the root node\n"); )
      struct btree_mem_node * n = bt->root;
      bt->root = bt->root->child[0];
      free(n);
      }
   return data;
   }

////////////////////////////////////////////////////////////////////////////////
// Assumption: node has bt->mindeg keys or is the root.

void *
btree_mem_node_remove(struct btree_mem * bt, struct btree_mem_node * node, void * key, void ** key_value, void ** data_value)
   {

   DEBUG_BLOCK( btree_mem_node_print_keys(node); )

   int i = node->nkeys - 1;
   while ( (i >= 0) && (bt->cmp(key, node->key[i]) < 0) )
      i--;

   if ( (i >= 0) && (bt->cmp(key, node->key[i]) == 0) )
      {
      // key is in node at key[i]
      *key_value = node->key[i];
      *data_value = node->data[i];
      //      printf("Will return key_value %p (%d)   data_value %p (%s)\n", *key_value, *((int*)(*key_value)), (char*)(*data_value));
      if (node->leaf)
	 {
	 // case 1
	 DEBUG_BLOCK( printf("remove: case 1\n"); )
	 btree_mem_node_remove_key(node, i);
	 return *data_value;
	 }
      else
	 {
	 struct btree_mem_node * prev = node->child[i];
	 struct btree_mem_node * next = node->child[i+1];

	 // if (child prev that precedes key in node has at least bt->mindeg keys) ||
	 //    (child next that follows  key in node has at least bt->mindeg keys)
	 if ( (prev->nkeys >= bt->mindeg) || (next->nkeys >= bt->mindeg) )
	    {
	    void * kp;
	    if (prev->nkeys >= bt->mindeg)
	       {
	       // case 2a
	       DEBUG_BLOCK( printf("remove: case 2a\n"); )
	       kp = btree_mem_node_last_key(prev);
	       //	       printf("remove: kp %d\n", *((int*)kp)); 
	       node->data[i] = btree_mem_node_remove(bt, prev, kp, key_value, data_value);
	       }
	    else
	       {
	       // case 2b
	       DEBUG_BLOCK( printf("remove: case 2b\n"); )
	       kp = btree_mem_node_first_key(next);
	       //	       printf("remove: kp %d\n", *((int*)kp)); 
	       node->data[i] = btree_mem_node_remove(bt, next, kp, key_value, data_value);
	       }
	    //	    printf("i %d\n", i);
	    //	       printf("remove: kp %d\n", *((int*)kp)); 
	    node->key[i] = kp;
	    //	    	    printf("=====\n"); 	    btree_mem_node_dump(node, printit, 0); 	    printf("=====\n");
	    //		    printf("returning %p\n", *data_value);
	    return *data_value;
	    }
	 else
	    {
	    // case 2c
	    DEBUG_BLOCK( printf("remove: case 2c\n"); )
	    btree_mem_node_add_key(bt, prev, node->key[i], node->data[i]);
	    btree_mem_node_remove_key(node, i);	
	    btree_mem_node_merge(prev, next);
	    return *data_value;
	    }
	 }
      }
   else
      {
      if (node->leaf)
	 {
	 DEBUG_BLOCK( printf("remove: key not found\n"); )
	 return NULL;
	 }

      // key is in child[i+1], if anywhere
      struct btree_mem_node * c = node->child[i+1];

      DEBUG_BLOCK( printf("remove: key should be in node %p\n", c); )
      if (c->nkeys == bt->mindeg - 1)
	 {
	 struct btree_mem_node * prev = NULL;
	 if (i >= 0)
	    prev = node->child[i];

	 struct btree_mem_node * next = NULL;
	 if (i+2 <= node->nkeys)
	    next = node->child[i+2];

	 DEBUG_BLOCK( printf("remove: prev %p  next %p\n", prev, next); )

	 if ( (prev) && (prev->nkeys >= bt->mindeg) )
	    {
	    // case 3a (with prev sibling)
	    // Move node->key[i] into c->key[0] shifting over existing keys
	    // Move prev->key[prev->nkeys-1] into node->key[i]
	    // if (prev->leaf)
	    //     move prev->child[prev->nkeys] into c->child[0] shifting over existing children
	    // prev->nkeys--;
	    
	    DEBUG_BLOCK( printf("remove: case 3a.1\n"); )

	    btree_mem_node_add_key(bt, c, node->key[i-1], node->data[i-1]);

	    node->key[i-1]  = prev->key[prev->nkeys-1];
	    node->data[i-1] = prev->data[prev->nkeys-1];

	    if (prev->leaf)
	       {
	       for (int j = c->nkeys; j >= 0; j--)
		  c->child[j+1] = c->child[j];
	       c->child[0] = prev->child[prev->nkeys];
	       }

	    prev->nkeys--;
	    }
	 else if ( (next) && (next->nkeys >= bt->mindeg) )
	    {
	    // case 3a (with next sibling)
	    // Symmetric with the above

	    DEBUG_BLOCK( printf("remove: case 3a.2\n"); )
	    btree_mem_node_add_key(bt, c, node->key[i+1], node->data[i+1]);

	    //	    printf("=====\n"); 	    btree_mem_node_dump(node, printit, 0); 	    printf("=====\n");

	    node->key[i+1]  = next->key[0];
	    node->data[i+1] = next->data[0];
	    for (int j = 0; j < next->nkeys - 1; j++)
	       {
	       next->key[j]  = next->key[j+1];
	       next->data[j] = next->data[j+1];
	       }

	    if (next->leaf)
	       {
	       c->child[c->nkeys] = next->child[0];
	       for (int j = 0; j < next->nkeys; j++)
		  next->child[j] = next->child[j+1];
	       }

	    next->nkeys--;

	    }
	 else
	    {
	    // case 3b
	    DEBUG_BLOCK( printf("remove: case 3b\n"); )
		  //	    printf("=====\n"); 	    btree_mem_node_dump(node, printit, 0); 	    printf("=====\n");
		  //	    printf("node %p  i %d\n", node, i); fflush(stdout);
		  //	    printf("adding %d to %p\n", *((int*)node->key[i+1]), (next ? next : prev));	    fflush(stdout);
	    btree_mem_node_add_key(bt, (next ? next : prev), node->key[i+1], node->data[i+1]);
	    //	    printf("=====\n");	    btree_mem_node_dump(node, printit, 0);	    printf("=====\n");

	    if (next)
	       {
	       //	       printf("here\n");
	       btree_mem_node_merge(c, next);
	       }
	    else
	       {
	       //	       printf("there\n");
	       btree_mem_node_merge(prev, c);
	       }

	    //	    printf("=====\n");	    btree_mem_node_dump(node, printit, 0);	    printf("=====\n");

	    btree_mem_node_remove_key(node, i+1);

	    if (next)
	       {
	       for (int j = i+2; j <= node->nkeys+1; j++)
		  node->child[j] = node->child[j+1];
	       }
	    else
	       {
	       for (int j = i; j <= node->nkeys+1; j++)
		  node->child[j] = node->child[j+1];
	       }

	    //	    printf("=====\n");	    btree_mem_node_dump(node, printit, 0);	    printf("=====\n");
	    }
	 }
      return btree_mem_node_remove(bt, c, key, key_value, data_value);
      }
   }

////////////////////////////////////////////////////////////////////////////////
// Assumption: node is not full
// Assumption: key is not already in node
void
btree_mem_node_add_key(struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data)
   {
   int i = node->nkeys;
   while ( (i > 0) && (bt->cmp(key, node->key[i-1]) < 0) )
      i--;

   for (int j = node->nkeys - 1; j >= i; j--)
      {
      node->key[j+1] = node->key[j];
      node->data[j+1] = node->data[j];
      }

   node->key[i] = key;
   node->data[i] = data;
   node->nkeys++;
   }

////////////////////////////////////////////////////////////////////////////////
// Assumption: node has at least mindeg keys.
// Does not remove any children.
void *
btree_mem_node_remove_key(struct btree_mem_node * node, int i)
   {
   void * data = node->key[i];
   for (int j = i; j < node->nkeys; j++)
      {
      node->key[j]  = node->key[j+1];
      node->data[j] = node->data[j+1];
      }
   node->nkeys--;
   return data;
   }

////////////////////////////////////////////////////////////////////////////////
// Return the lowest valued key of the subtree rooted at node
void *
btree_mem_node_first_key(struct btree_mem_node * node)
   {
   if (0 == node->nkeys)
      return NULL;

   if (node->leaf)
      return node->key[0];

   return btree_mem_node_last_key(node->child[0]);
   }

////////////////////////////////////////////////////////////////////////////////
// Return the highest valued key of the subtree rooted at node
void *
btree_mem_node_last_key(struct btree_mem_node * node)
   {
   if (0 == node->nkeys)
      return NULL;

   if (node->leaf)
      return node->key[node->nkeys-1];

   return btree_mem_node_last_key(node->child[node->nkeys]);
   }

/////////////////////////////////////////////////////////////////////////////////
// move all keys/data from node b into node a
// Assumption: all keys in b come after the keys in a
void
btree_mem_node_merge(struct btree_mem_node * a, struct btree_mem_node * b)
   {

   DEBUG_BLOCK( printf("merging %p with %p\n", a, b); )
   int i = a->nkeys;
   int j = 0;
   for (;  j < b->nkeys; i++, j++)
      {
      a->key[i] = b->key[j];
      a->data[i] = b->data[j];
      }
   a->nkeys += b->nkeys;
   b->nkeys = 0;
   }

/////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
void
btree_mem_node_print_keys(struct btree_mem_node * node)
   {
   printf("node: %p nkeys %10d keys: ", node, node->nkeys);
   for (int i = 0; i < node->nkeys; i++)
      printf("%d ", *((int *) node->key[i]));
   printf("\n");
   }
#endif

