#include <ctype.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>

#include "list.h"

struct trie_node
   {
   char key;
   struct trie_node * parent;
   struct trie_node * child;
   struct trie_node * next_sibling;
   struct trie_node * prev_sibling;
   void * data;
   };

struct trie
   {
   struct trie_node * root;
   size_t size;
   size_t count;
   };


static struct list * subkey_list = NULL;
static size_t subkey_list_limit = 0;

struct trie_node * trie_search_children(struct trie_node * node, char k);
void trie_visit_node(struct trie * trie, struct trie_node * node, void (*op)(char * k, void * d), char * key);
void trie_print_node(struct trie_node * node);
void trie_print_node(struct trie_node * node);
void trie_node_dump_preorder(struct trie * trie, struct trie_node * node, int depth);
void trie_node_dump_raw(struct trie_node * node, int depth);
void trie_subtree_list_push(char * k, void * d);
struct trie_node * trie_find_node(struct trie * trie, char * key);

////////////////////////////////////////////////////////////////////////////////
struct trie *
trie_new()
   {
   struct trie * trie = malloc(sizeof(struct trie));
   if (NULL == trie)
      return NULL;

   trie->root = malloc(sizeof(struct trie_node));
   if (NULL == trie->root)
      return NULL;

   trie->root->parent = trie->root->child = trie->root->next_sibling = NULL;
   trie->root->key = '\0';
   trie->root->data = NULL;
   trie->size = trie->count = 0;

   return trie;
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_delete(struct trie * trie)
   {
   if (NULL == trie)
      return;

   free(trie);
   }

////////////////////////////////////////////////////////////////////////////////
size_t
trie_size(struct trie * trie)
   {
   if (NULL == trie)
      return 0;
   return trie->size;
   }

////////////////////////////////////////////////////////////////////////////////
size_t
trie_count(struct trie * trie)
   {
   if (NULL == trie)
      return 0;
   return trie->count;
   }

////////////////////////////////////////////////////////////////////////////////
char *
trie_insert(struct trie * trie, char * key, void * data)
   {
   if ( (NULL == trie) || (NULL == key) )
      return NULL;

   struct trie_node * curr = trie->root;
   struct trie_node * result;

   size_t len = strlen(key);
   for (int i = 0; i <= len; ++i)
      {
      result = trie_search_children(curr, key[i]);
      if (NULL == result)
	 {
	 // current char is not in the trie yet, allocate a new node and attach it to the children
	 struct trie_node * new = malloc(sizeof(struct trie_node));
	 if (NULL == new)
	    return NULL;
	 new->key = tolower(key[i]);
	 new->child = new->next_sibling = new->prev_sibling = NULL;
	 new->data = NULL;

	 if (NULL == curr->child)
	    curr->child = new;
	 else
	    {
	    struct trie_node * p = curr->child;
	    while (p->next_sibling)
	       p = p->next_sibling;
	    new->prev_sibling = p;
	    p->next_sibling = new;
	    }

	 new->parent = curr;
	 curr = new;
	 trie->size++;
	 }
      else
	 curr = result;
      }

   // At this point, curr is either pointing to a new \0 node with a NULL data pointer,
   // or an existing \0 node with a valid data pointer (in which case we were asked to "insert" 
   // a key that was already present in the trie.
   // For now I am simply ignoring collisions.  We will update the data.

   curr->data = data;

   trie->count++;

   return key;
   }

////////////////////////////////////////////////////////////////////////////////
struct trie_node *
trie_search_children(struct trie_node * node, char k)
   {
   struct trie_node * p = node->child;
   if (NULL == p)
      return NULL;

   while ( (NULL != p) && (p->key != tolower(k)) )
      p = p->next_sibling;
   return p;
   }

////////////////////////////////////////////////////////////////////////////////
struct trie_node *
trie_find_node(struct trie * trie, char * key)
   {
   if ( (NULL == trie) || (NULL == key) )
      return NULL;

   struct trie_node * node = trie->root;
   size_t len = strlen(key);
   for (int i = 0; i <= len; i++)
      {
      node = trie_search_children(node, key[i]);
      if (NULL == node)
	 return NULL;
      if ('\0' == key[i])
	 return node;
      }

   // Should never get here
   return NULL;
   }

////////////////////////////////////////////////////////////////////////////////
void *
trie_find(struct trie * trie, char * key)
   {
   struct trie_node * node = trie_find_node(trie, key);
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
void *
trie_remove(struct trie * trie, char * key)
   {
   if ( (NULL == trie) || (NULL == key) )
      return NULL;

   struct trie_node * node = trie_find_node(trie, key);
   if (NULL == node)
      return NULL;

   void * data = node->data;

   while ( (NULL == node->prev_sibling) && (NULL == node->prev_sibling) ) 
      {
      struct trie_node * p = node;
      node = node->parent;
      free(p);
      trie->size--;
      }

   if (node == trie->root)
      return data;

   if (node->prev_sibling)
      node->prev_sibling->next_sibling = node->next_sibling;
   else
      node->parent->child = node->next_sibling;

   if (node->next_sibling)
      node->next_sibling->prev_sibling = node->prev_sibling;

   free(node);

   trie->count--;

   return data;
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_walk_keys(struct trie * trie, void (*op)(char * k, void * d))
   {
   if (NULL == trie)
      return;
   trie_visit_node(trie, trie->root, op, "");
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_visit_node(struct trie * trie, struct trie_node * node, void (*op)(char * k, void * d), char * key)
   {
   if ( (node->key == '\0') && (node != trie->root) )
      op(key, node->data);
   else
      {
      size_t len = strlen(key);
      char * k = malloc(len+2);
      if (NULL == k)
	 return;
      strcpy(k, key);
      k[len] = node->key;
      k[len+1] = '\0';
      struct trie_node * p = node->child;
      while (NULL != p)
	 {
	 trie_visit_node(trie, p, op, k);
	 p = p->next_sibling;
	 }
      free(k);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void trie_dump_raw(struct trie * trie)
   { 
   if (NULL == trie)
      return;
  
   int depth = -1;
   trie_node_dump_raw(trie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_node_dump_raw(struct trie_node * node, int depth)
   {
   for (int i=0; i < depth; i++)
      printf(" ");
   trie_print_node(node);
   if (node->child)
      trie_node_dump_raw(node->child, depth+1); 
   if (NULL != node->next_sibling)
      trie_node_dump_raw(node->next_sibling, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
void trie_dump(struct trie * trie)
   { 
   if (NULL == trie)
      return;
  
   int depth = -1;
   trie_node_dump_preorder(trie, trie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_node_dump_preorder(struct trie * trie, struct trie_node * node, int depth)
   {
   printf("%c", node->key);
   if ( (node != trie->root) && ('\0' == node->key) )
      printf("\n");
   if (node->child)
      trie_node_dump_preorder(trie, node->child, depth+1);
   if (NULL != node->next_sibling)
      {
      for (int i=0; i < depth; i++)
	 printf(" ");
      trie_node_dump_preorder(trie, node->next_sibling, depth);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_print_node(struct trie_node * node)
   {
   printf("%p\tchild %p\tnext_sibling %p\tkey %c\tdata %p\n", (void *) node, (void *) node->child, 
	  (void *) node->next_sibling, node->key, node->data);
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_subtree_list_push(char * k, void * d)
   {
   if (NULL == subkey_list)
      return;

   if ( (0 == subkey_list_limit) || (list_size(subkey_list) < subkey_list_limit) )
      list_push_front(subkey_list, k);
   }

////////////////////////////////////////////////////////////////////////////////
// Not reentrant since it uses the package static subkeys_list variable.
// However, that variable is only used while building a subkey list, so you can
// get more than one subkey if you want.  Just don't try to get them simultaneously.
struct list *
trie_get_subkeys(struct trie * trie, char * prefix, size_t limit)
   {
   if ( (NULL == trie) || (NULL == prefix) )
      return NULL;

   struct trie_node * node = trie_find_node(trie, prefix);
   if (NULL == node)
      return NULL;

   subkey_list = list_new();
   if (NULL == subkey_list)
      return NULL;
   subkey_list_limit = limit;

   size_t len = strlen(prefix);
   char * pref = malloc(len+1);
   if (NULL == pref)
      return NULL;
   strncpy(pref, prefix, len-1);
   pref[len-1] = '\0';

   trie_visit_node(trie, node->parent, trie_subtree_list_push, pref);
   free(pref);

   return subkey_list;
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_free_subkey_list(struct list * list)
   {
   if (NULL == list)
      return;

   struct list_iterator * it = list_begin(list);
   void * data;
   while (data = list_remove(it))
      free(data);

   list_delete(list);
   }
